/**
 * @file gif.h
 * GIF loading handling.
 *
 * Description of gif.h
 *
 * @addtogroup gif GIF file handling
 * @{
 *
 */
		
#include "ln.h"
		
#ifndef __GIF_H__
#define __GIF_H__
		
// ----------------- Types		-----------------
		
/**
 * GIF image mode
 */
typedef enum
{
	GIFMODE_BITMAP,		/**< GIF image data is bitmapped */
	GIFMODE_TILES,		/**< GIF image data is tiled */
		
} gifMode;
		
/**
 * GIF handler
 */
typedef struct 
{
	u32 dimX;			/**< GIF image width in pixels */
	u32 dimY;			/**< GIF image height in pixels */
	u32 frmCount;		/**< Number of frames if animation presents (1 for no animation) */
	u16 loopCnt;		/**< GIF Loop count (0 for unlimited)*/
	u16 colCount;		/**< Number of colors the GIF posses, length of gifPalette */
	gifMode dataMode;	/**< GIF frame mode */
	u8* gifFrames;		/**< GIF bitmap buffer, as linear animation */
	u16* gifPalette;	/**< GIF palette */
	u32* gifTiming;		/**< GIF Timing indexes per frame */
		
} tGifHandler, *pGifHandler;
		
// ----------------- Functions	-----------------
		
/**
 * Fills the handler with GIF information
 * 
 * @param filename GIF File name to be loaded
 * @return Inited GIF Handler or NULL for error
 */
pGifHandler gifHandlerLoad(const char* filename);
		
/**
 * Removes all data used by the GIF handler
 * 
 * @param gifHandler GIF Handler to be destroyed
 * @note This also clears palette and bitmap(s)
 */
void gifHandlerDestroy(pGifHandler gifHandler);
		
/**
 * Covnerts the bitmap(s) in the GIF handler to tiles
 * 
 * @param gifHandler GIF Handler to be converted
 * @warning This is irreversible!
 */
bool gifConvertTiles(pGifHandler gifHandler);
		
#endif
		
/**
 * @}
 */
